"use strict";
// NOTE: Firebase doesn't support complex quering like most DBs (
// https://firebase.google.com/docs/firestore/query-data/queries). Therefore, we will do all the
// querying in memory. In the future we may want to support a minimal set of queries against the
// firestore DB so that we have the option of working with larger datasets.
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __rest = (this && this.__rest) || function (s, e) {
    var t = {};
    for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p) && e.indexOf(p) < 0)
        t[p] = s[p];
    if (s != null && typeof Object.getOwnPropertySymbols === "function")
        for (var i = 0, p = Object.getOwnPropertySymbols(s); i < p.length; i++) {
            if (e.indexOf(p[i]) < 0 && Object.prototype.propertyIsEnumerable.call(s, p[i]))
                t[p[i]] = s[p[i]];
        }
    return t;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const memory_store_1 = __importDefault(require("./memory-store"));
const deep_clone_1 = require("../utils/deep-clone");
const reorder_1 = require("./reorder");
class FirebaseStore extends memory_store_1.default {
    constructor() {
        super(...arguments);
        this.className = 'FirebaseStore';
        this._onReorder = (id, doc) => __awaiter(this, void 0, void 0, function* () {
            // Update the order in firebase of the affected doc
            yield this._docSet({
                id,
                doc,
                options: {
                    merge: true, // allow for partial updates
                },
            });
        });
    }
    // React to real-time changes made remotely
    _listenToChanges() {
        return __awaiter(this, void 0, void 0, function* () {
            // We order by createdAt as per https://firebase.google.com/docs/firestore/manage-data/add-data,
            // "Firestore auto-generated IDs do not provide any automatic ordering. If you want to be able
            // to order your documents by creation date, you should store a timestamp as a field in the
            // documents." We can only order by a single attribute (unless we create an index) so we'll pick
            // the order.
            this._coll.orderBy('createdAt').onSnapshot((snapshot) => {
                snapshot.docChanges().forEach((change) => {
                    const data = change.doc.data();
                    if (change.type === 'removed') {
                        this._docs.delete(data.id);
                    }
                    else {
                        this._docs.set(data.id, data);
                    }
                });
                // First load?
                if (!this.isLoaded()) {
                    // Emit on next tick so that caller has a change to listen
                    setTimeout(() => {
                        this._emitDidLoad();
                    });
                }
            }, (err) => {
                this._emitError(err);
            });
        });
    }
    // async _loadDocs() {
    //   const docs = await this._coll.get();
    //   docs.forEach(doc => {
    //     const doc = doc.data();
    //     this._docs.set(doc.id, doc);
    //   });
    // }
    _getFirebase(props) {
        return props.firebase ? props.firebase : this._global.firebase;
    }
    _initializeApp(props) {
        // Has Firebase been initialized? We must only do this once per app.
        if (!this._fb.apps.length) {
            this._fb.initializeApp({
                apiKey: props.apiKey,
                authDomain: props.authDomain,
                projectId: props.projectId,
            });
        }
    }
    _init(props) {
        return __awaiter(this, void 0, void 0, function* () {
            this._fb = this._getFirebase(props);
            this._initializeApp(props);
            this._db = this._fb.firestore();
            // Disable warning
            const settings = { timestampsInSnapshots: true };
            this._db.settings(settings);
            this._coll = this._db.collection(props.collection);
            this._listenToChanges();
            // Not needed as we get the docs by listening to the changes
            // await this._loadDocs();
        });
    }
    create(props) {
        super.create(Object.assign({}, props, {
            // Mute didLoad in the base classes as we need to emit didLoad after all the docs have been
            // loaded asynchronously
            muteDidLoad: true,
        }));
        // For mocking
        this._global = global;
        this.set({
            schema: {
                component: 'Form',
                fields: [
                    {
                        // For mocking
                        name: 'firebase',
                        component: 'Field',
                    },
                    {
                        name: 'apiKey',
                        component: 'TextField',
                        required: true,
                    },
                    {
                        name: 'authDomain',
                        component: 'TextField',
                        required: true,
                    },
                    {
                        name: 'projectId',
                        component: 'TextField',
                        required: true,
                    },
                    {
                        name: 'collection',
                        component: 'TextField',
                        required: true,
                    },
                ],
            },
        });
        // Don't actually connect to Firebase unless we specify an API Key
        if (props.apiKey !== undefined) {
            this._init(props);
        }
    }
    _docSet({ id, doc, options }) {
        return __awaiter(this, void 0, void 0, function* () {
            // Don't actually connect to Firebase unless we specify an API Key
            if (this.get('apiKey')) {
                // The Firebase API is not capable of storing undefined values so we remove them
                doc = JSON.parse(JSON.stringify(doc));
                // Update the Firebase store
                return this._coll.doc(id).set(doc, options);
            }
        });
    }
    _createDoc({ fieldValues, id, order, reorder }) {
        return __awaiter(this, void 0, void 0, function* () {
            if (this._shouldSetToLastOrder(order, reorder)) {
                order = this._numUnarchived();
            }
            const userId = this._getUserId();
            const doc = this._buildDoc({ fieldValues, id, order, userId });
            // Note: we need to update the underlying MemoryStore so that the data is there after this
            // function completes even though Firebase will emit an onWrite with the changed data.
            this._docs.set(doc.id, doc, undefined, undefined, true, this._onReorder);
            yield this._docSet({
                id: doc.id,
                doc,
                order: doc.order,
            });
            return doc;
        });
    }
    _modifyDoc({ id, fieldValues, archivedAt, order }) {
        return __awaiter(this, void 0, void 0, function* () {
            // We use cloneDeep so that we don't modify the data in the memory store before we modify the
            // data in the Firebase store. The write to Firebase could always fail.
            let doc = (0, deep_clone_1.cloneDeep)(this._docs.get(id));
            doc = this._setDoc({
                doc,
                fieldValues,
                archivedAt,
                order,
            });
            // Note: we need to update the underlying MemoryStore so that the data is there after this
            // function completes even though Firebase will emit an onWrite with the changed data.
            this._docs.set(doc.id, doc, undefined, undefined, true, this._onReorder);
            // The value has to be set in Firebase after it is set in the mapa so that _docs.set() can
            // detect when there is a change to the order
            yield this._docSet({
                id: doc.id,
                doc,
                options: {
                    merge: true, // allow for partial updates
                },
            });
            return doc;
        });
    }
    _updateDoc(props) {
        return __awaiter(this, void 0, void 0, function* () {
            return this._modifyDoc(props);
        });
    }
    _archiveDoc(props) {
        return __awaiter(this, void 0, void 0, function* () {
            return this._modifyDoc(Object.assign(Object.assign({}, props), { archivedAt: this._now(), order: reorder_1.Reorder.DEFAULT_ORDER }));
        });
    }
    _restoreDoc(_a) {
        var { reorder, order } = _a, props = __rest(_a, ["reorder", "order"]);
        return __awaiter(this, void 0, void 0, function* () {
            if (this._shouldSetToLastOrder(order, reorder)) {
                order = this._numUnarchived();
            }
            return this._modifyDoc(Object.assign(Object.assign({}, props), { archivedAt: null, order }));
        });
    }
    _waitForDataToBeLoaded() {
        return __awaiter(this, void 0, void 0, function* () {
            // Wait until the data is first loaded before trying to get it from the underlying MemoryStore
            yield this.resolveAfterLoad();
        });
    }
    _getDoc(props) {
        const _super = Object.create(null, {
            _getDoc: { get: () => super._getDoc }
        });
        return __awaiter(this, void 0, void 0, function* () {
            yield this._waitForDataToBeLoaded();
            return _super._getDoc.call(this, props);
        });
    }
    _getAllDocs(props) {
        const _super = Object.create(null, {
            _getAllDocs: { get: () => super._getAllDocs }
        });
        return __awaiter(this, void 0, void 0, function* () {
            yield this._waitForDataToBeLoaded();
            return _super._getAllDocs.call(this, props);
        });
    }
}
exports.default = FirebaseStore;
