"use strict";
// Timestamps are set in the store as some stores, e.g. those backed by remote databases, may wish
// to manage the timestamps for better reliability.
//
// Structure of doc in a store, e.g.
//  {
//     id,
//     userId,
//     ...,
//     fieldValues: {
//       firstName,
//       lastName,
//       ...
//     }
//   }
// In other words, we nest the default fields in the root of the document and then nest the
// non-default fields in fieldValues. This is done because:
//   1. There are some properties, like the cursor, which should not be a field as they shouldn't be
//      displayed to the user. However, these fields need to be associated with the form.
//   2. It allows for the creation of extra properties in the future without adding them to the form
//      as default fields
//   3. Default fields are not included in the fieldValues as this allows the default fields,
//      especially the id to be at the root of the record and not nested in the fieldValues. We also
//      don’t want to duplicate these default fields in both the root of the record and in the field
//      values as this can waste memory when storing the data
//
// We use a numeric order attribute to order docs in a set, e.g. allowing for drag-to-order UIs. The
// downside to this approach, is that moving a doc requires updating the order of all subsequent
// docs (https://dba.stackexchange.com/q/36875/94046). A previous design used a beforeId, similar to
// the beforeKey construct in the Mapa, and this only required at most 2 writes to move a doc. The
// downside; however, was that stores can return data in any order, or there can be race conditions,
// which result in beforeIds for docs that don't exist. This problematic beforeIds cause a great
// deal of extra complexity.
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const ui_component_1 = __importDefault(require("../ui-component"));
const access_1 = __importDefault(require("../access"));
const utils_1 = __importDefault(require("../utils/utils"));
const uber_utils_1 = __importDefault(require("../utils/uber-utils"));
const date_field_1 = __importDefault(require("../fields/date-field"));
const reorder_1 = require("./reorder");
const globals_1 = __importDefault(require("../globals"));
const registrar_1 = __importDefault(require("../compiler/registrar"));
class Store extends ui_component_1.default {
    constructor() {
        super(...arguments);
        this.className = 'Store';
    }
    create(props) {
        super.create(props);
        this.set({
            schema: {
                component: 'Form',
                fields: [
                    {
                        name: 'where',
                        component: 'Field',
                    },
                ],
            },
        });
        // For mocking
        this._access = access_1.default;
        this._globals = globals_1.default;
        this._registrar = registrar_1.default;
        if (!props || !props.muteDidLoad) {
            this._emitDidLoad();
        }
    }
    _tryAndHandleError(promiseFactory) {
        return __awaiter(this, void 0, void 0, function* () {
            return uber_utils_1.default.tryAndDisplayErrorIfAPIError(promiseFactory);
        });
    }
    _shouldSetToLastOrder(order, reorder) {
        return (order === null || order === undefined) && reorder;
    }
    createDoc(props) {
        return __awaiter(this, void 0, void 0, function* () {
            return this._tryAndHandleError(() => {
                // Omit values based on access
                const fieldValues = this._access.valuesCanCreate(props.form, {
                    default: false,
                });
                const id = props.form.getValue('id');
                let order = props.form.getValue('order');
                if ((order === null || order === undefined) && !props.reorder) {
                    order = reorder_1.Reorder.DEFAULT_ORDER;
                }
                return this._createDoc(Object.assign(Object.assign({}, props), { fieldValues, id, order }));
            });
        });
    }
    getDoc(props) {
        return __awaiter(this, void 0, void 0, function* () {
            return this._tryAndHandleError(() => __awaiter(this, void 0, void 0, function* () {
                if (props.id) {
                    return this._getDoc(props);
                }
                else {
                    // Reuse the logic in _getAllDocs to query with "where" and return the first doc
                    const docs = yield this._getAllDocs(Object.assign(Object.assign({}, props), { showArchived: null }));
                    return docs.edges.length > 0 ? docs.edges[0].node : null;
                }
            }));
        });
    }
    getAllDocs(props) {
        return __awaiter(this, void 0, void 0, function* () {
            return this._tryAndHandleError(() => {
                return this._getAllDocs(props);
            });
        });
    }
    _upsertDoc(props) {
        return __awaiter(this, void 0, void 0, function* () {
            const id = props.id;
            let exists = false;
            if (id) {
                exists = yield this._hasDoc({ id });
            }
            if (exists) {
                return this.updateDoc(props);
            }
            else {
                return this.createDoc(props);
            }
        });
    }
    upsertDoc(props) {
        return __awaiter(this, void 0, void 0, function* () {
            return this._tryAndHandleError(() => {
                const id = props.form.getValue('id');
                return this._upsertDoc(Object.assign(Object.assign({}, props), { id }));
            });
        });
    }
    updateDoc(props) {
        return __awaiter(this, void 0, void 0, function* () {
            return this._tryAndHandleError(() => {
                // Omit values based on access
                const fieldValues = this._access.valuesCanUpdate(props.form, {
                    default: false,
                });
                const id = props.form.getValue('id');
                const order = props.form.getValue('order');
                return this._updateDoc(Object.assign(Object.assign({}, props), { fieldValues, id, order }));
            });
        });
    }
    archiveDoc(props) {
        return __awaiter(this, void 0, void 0, function* () {
            return this._tryAndHandleError(() => {
                return this._archiveDoc(props);
            });
        });
    }
    restoreDoc(props) {
        return __awaiter(this, void 0, void 0, function* () {
            return this._tryAndHandleError(() => {
                return this._restoreDoc(props);
            });
        });
    }
    _emitError(err) {
        this.emitChange('err', err);
    }
    _now() {
        // We use a DateField to avoid Firestore's automatic conversion of Dates to Firebase style
        // timestamps.
        const date = new date_field_1.default({ now: true });
        return date.getValue();
    }
    _buildDoc({ fieldValues, id, userId, order }) {
        const createdAt = this._now();
        return {
            id: id ? id : utils_1.default.uuid(),
            archivedAt: null,
            createdAt,
            updatedAt: createdAt,
            userId: userId ? userId : null,
            order,
            fieldValues,
        };
    }
    _setDoc({ doc, fieldValues, archivedAt, order }) {
        if (fieldValues !== undefined) {
            // Merge so that we support partial updates
            doc.fieldValues = Object.assign(doc.fieldValues, fieldValues);
        }
        if (archivedAt !== undefined) {
            doc.archivedAt = archivedAt;
        }
        doc.updatedAt = this._now();
        if (order === undefined) {
            // Set to null instead of setting to undefined as stores like Firebase don't support undefined
            // values
            doc.order = reorder_1.Reorder.DEFAULT_ORDER;
        }
        else {
            doc.order = order;
        }
        return doc;
    }
    _getUserId() {
        if (this._registrar.client) {
            const session = this._registrar.client.user.getSession();
            if (session) {
                return session.user.id;
            }
        }
    }
}
exports.default = Store;
